from pyonepassword.string import RedactedString


def test_redacted_string_010():
    """
    Create a RedactedString object

    Verify the redacted string matches the expected value
    """
    original_string = "An example, unredacted string"
    expected = "************************tring"

    redacted = RedactedString(original_string)
    assert expected == str(redacted)


def test_redacted_string_020():
    """
    Create a RedactedString object with an arbitrarly long unmask_len

    Verify:
        The unmasked portion of the redacted string is appropriate clamped
        to 20% of the string length
    """
    original_string = "An example, unredacted string"
    expected = "************************tring"

    # the max unmasked length of the above string be clamped to 5
    long_unmask_len = 12

    redacted = RedactedString(original_string, unmask_len=long_unmask_len)
    assert expected == str(redacted)


def test_redacted_string_030():
    """
    Create a RedactedString object with an short unmask_len

    Verify:
        The unmasked portion of the redacted string is appropriate restricted
        to the specified value
    """
    original_string = "at 60 characters. this is an even longer string than normal."
    expected = "*************************************************han normal."

    unmask_len = 11

    redacted = RedactedString(original_string, unmask_len=unmask_len)
    assert expected == str(redacted)


def test_redacted_string_040():
    """
    Create a RedactedString object with an arbitrarly long unmask_len

    Verify:
        The unmasked portion of the redacted string is appropriate clamped
        to 20% of the string length
    """
    original_string = "at 60 characters. this is an even longer string than normal."
    # the max unmasked length of the above string be clamped to 12
    expected = "************************************************than normal."

    long_unmask_len = 17

    redacted = RedactedString(original_string, unmask_len=long_unmask_len)
    assert expected == str(redacted)


def test_redacted_string_050():
    """
    Create a RedactedString object with a negative unmask_len

    Verify:
        The unmasked portion of the redacted string is the default length
    """
    original_string = "at 60 characters. this is an even longer string than normal."
    expected = "*******************************************************rmal."

    # a negative unmask length should result in the default unmask length
    long_unmask_len = -5

    redacted = RedactedString(original_string, unmask_len=long_unmask_len)
    assert expected == str(redacted)


def test_redacted_string_060():
    """
    Create a RedactedString object from a short string

    Verify:
        The entire string is masked
    """
    original_string = "short"

    # a very short string should remain completely masked
    expected = "*****"

    redacted = RedactedString(original_string)
    assert expected == str(redacted)


def test_redacted_string_070():
    """
    Create a RedactedString object from a short string,
        specifying a unmask_len of 20% of the string length

    Verify:
        The unmask_len is ignored, and entire string is masked
    """
    original_string = "short"

    # even when specifying an unmask_len,
    # a very short string should remain completely masked
    expected = "*****"

    unmask_len = 1

    redacted = RedactedString(original_string, unmask_len=unmask_len)
    assert expected == str(redacted)


def test_redacted_string_090():
    """
    Create a RedactedString from an existing RedactedString object

    Verify:
        - The two objects are equal
        - The second redacted value matches the expected value
    """
    original_string = "An example, unredacted string"
    expected = "************************tring"

    r1 = RedactedString(original_string)
    r2 = RedactedString(r1)
    assert expected == str(r2)
    assert r2 == r1


def test_redacted_string_100():
    """
    Create a RedactedString object, and concatenate a second string to it

    Verify the redacted value matches the expected value
    """
    original_string = "An example, unredacted string"

    expected_redacted = "*****************************************onger"

    redacted = RedactedString(original_string)
    redacted += ", now even longer"
    assert expected_redacted == str(redacted)


def test_redacted_string_110():
    """
    Create a RedactedString object, and concatenate a second string to it

    Verify the object is equal to the expected unredacted string
    """
    original_string = "An example, unredacted string"
    expected_original = "An example, unredacted string, now even longer"

    redacted = RedactedString(original_string)
    redacted += ", now even longer"
    assert expected_original == redacted


def test_redacted_string_120():
    """
    Create:
        - a RedactedString object,
        - a second RedactedString object by adding a second string to the first

    Verify the second redacted value matches the expected value
    """
    original_string = "An example, unredacted string"
    expected_redacted = "*****************************************onger"

    r1 = RedactedString(original_string)
    r2 = r1 + ", now even longer"
    assert expected_redacted == str(r2)


def test_redacted_string_130():
    """
    Create:
        - Two RedactedString objects
        - A third RedactedString object by adding the first two together

    Verify the third redacted value matches the expected value
    """
    orig_1 = "Redacted part 1."
    orig_2 = " Redacted part 2."
    expected_redacted = "****************************rt 2."

    r1 = RedactedString(orig_1)
    r2 = RedactedString(orig_2)
    r3 = r1 + r2
    assert expected_redacted == str(r3)


def test_redacted_string_140():
    """
    Create:
        - Two RedactedString objects
        - A third RedactedString object by adding the first two together

    Verify the third's 'original' property matches the expected original
    """
    orig_1 = "Redacted part 1."
    orig_2 = " Redacted part 2."

    expected_original = orig_1 + orig_2

    r1 = RedactedString(orig_1)
    r2 = RedactedString(orig_2)
    r3 = r1 + r2
    assert expected_original == r3.original


def test_redacted_string_150():
    """
    Create:
        - Two RedactedString objects
        - A third RedactedString object by adding the first two together

    Verify an abitrary slice of the third object matches the same arbitrary slice of the original
    """
    orig_1 = "Redacted part 1."
    orig_2 = " Redacted part 2."

    expected_original = orig_1 + orig_2

    r1 = RedactedString(orig_1)
    r2 = RedactedString(orig_2)
    r3 = r1 + r2

    assert expected_original[:5] == r3[:5]



import json
from typing import List, Union

from ..json import safe_unjson
from ._item_descriptor_registry import OPItemDescriptorFactory
from .item_types._item_descriptor_base import OPAbstractItemDescriptor


class OPItemList(List[OPAbstractItemDescriptor]):
    def __init__(self, item_list_json: Union[str, List], generic_okay=False):
        super().__init__()
        item_list = safe_unjson(item_list_json)
        for i_dict in item_list:
            descriptor = OPItemDescriptorFactory.item_descriptor(
                i_dict, generic_okay=generic_okay)
            self.append(descriptor)

        # 'op item list' returns items in a non-deterministic order
        # so sorting ourselves hopefully ensures we are in a consistent order every time
        # this will mainly make it easier for testing
        # so multiple queries for the same list of items will yield the same result
        self.sort()

    def serialize(self, indent=None) -> str:
        json_str = json.dumps(self, indent=indent)
        return json_str

    def sort(self):
        # we really want to be sorted by title, since that's the most
        # human friendly (mainly for troubleshooting)
        # but in the case if title collisions, this is still non-deterministic
        # since 'op' may return them in a non-deterministic order
        # since list.sort() is considered stable we can ensure items
        # with the same title are in a particular order relative to each other by...
        # first, sorting by unique id
        super().sort(key=lambda item: item.unique_id)
        # Then sort by title
        super().sort(key=lambda item: item.title)



import os

from setuptools import find_packages, setup
from setuptools.command.egg_info import egg_info


def project_path():
    path = os.path.dirname(os.path.abspath(__file__))
    return path


old_cwd = os.getcwd()
proj_path = project_path()
os.chdir(proj_path)

about = {}

with open(os.path.join(proj_path, "pyonepassword", "__about__.py"), "r") as fp:
    exec(fp.read(), about)

with open(os.path.join(proj_path, "README.md"), "r") as fp:
    long_description = fp.read()


packages = find_packages(
    where=".", include=["pyonepassword", "pyonepassword.*"])


# Stale *.egg-info items don't get cleaned automatically
# and can result in things being packaged that shouldn't be or not packaged
# that should be
class CleanEggInfoCommand(egg_info):
    def run(self):
        import glob
        import shutil
        rm_list = glob.glob('*.egg-info')
        print("egg_info: removing egg-info")
        for rm_path in rm_list:
            print("rm_path")
            shutil.rmtree(rm_path)
        super().run()


setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__summary__"],
    license="MIT",
    author="Zachary Cutlip",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zcutlip/pyonepassword",
    packages=packages,

    # We need python3.9 in order to use importlib.resources.files in templates.py
    python_requires='>=3.8',
    install_requires=[
        # importlib.resources.files requires python >=3.9
        # if python 3.8, need to install 3rd importlib-resources
        "importlib-resources>=5.2.0; python_version<'3.9'",
        "python-singleton-metaclasses"
    ],
    package_data={'pyonepassword': ['data/**', 'py.typed']},
    entry_points={"console_scripts":
                  ["opconfig=pyonepassword.opconfig_main:main"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    cmdclass={
        # TODO: this breaks building in tox>=4.0
        # disabling for now
        # 'egg_info': CleanEggInfoCommand
    },
)

os.chdir(old_cwd)



import json
import os
import tempfile
from typing import Dict, List, Optional

from ..py_op_exceptions import OPInvalidItemException
from ._new_field_registry import OPNewItemField, OPNewItemFieldFactory
from .fields_sections.item_field_base import OPItemField
from .fields_sections.item_section import (
    OPSection,
    OPSectionCollisionException
)
from .template_directory import OPTemplateDirectory
from .uuid import OPUniqueIdentifierBase32, is_uuid


class OPNewItemDataCollisionException(Exception):
    pass


class OPNewSection(OPSection):
    """
    A class for creating new sections for use with new items

    New sections may be created directly, or from existing sections
    """

    def __init__(self, section_label: str, section_id: Optional[str] = None):
        """
        Create a new section object

        Parameters
        ----------
        section_label: str
            The user-visible name of the section
        section_id: str, optional
            The unique identifier for this section. If none is provided, a random one will be generated
        """
        if not section_id:
            unique_id = OPUniqueIdentifierBase32()
            section_id = str(unique_id)
        section_dict = {
            "id": section_id,
            "label": section_label
        }
        super().__init__(section_dict)

    @classmethod
    def from_section(cls, section: OPSection):
        """
        Create a new section from an existing one

        If the existing section's ID is a random ID, a new random ID will be generated

        Parameters
        ----------
        section: OPSection
            The section object to duplicate

        Returns:
        new_section: OPNewSection
            The newly created section object
        """
        section_id = section.section_id
        if is_uuid(section_id):
            section_id = str(OPUniqueIdentifierBase32())
        label = section.label
        return cls(label, section_id=section_id)


class OPNewItemMixin:
    """
    A class that, when subclassed along with one of the item classes (OPLogin, etc.), allows a
    item of that type to be created from a template
    E.g.,
    class OPLoginItemTemplate(OPNewItemMixin, OPLoginItem):
        ...

    NOTE: It is essential OPNewItemMixin be named first, so its `__init_()` gets called first
    """
    FROM_TEMPLATE = True

    # Whether this item type can have a password generated
    # override for item types that do support passwords (e.g., Login)
    PASSWORDS_SUPPORTED = False

    def __init__(self, title: str, fields: List[OPItemField] = [], sections: List[OPSection] = [], tags: List[str] = [], extra_data={}):
        """
        Create an OPNewItemMixin object that can be used to create a new item entry

        Parameters
        ----------
        Title : str
            User viewable name of the item to create
        fields: List[OPItemField], optional
            List of OPItemField objects to associate with the item.
            NOTE: If the fields are from an exisiting item, and the field IDs are UUIDs, the field IDs will be regenerated
        sections: List[OPSection], optional
            List of OPSection objects to associate with the item.
            NOTE: If the sections are from an exisiting item, and the section IDs are UUIDs, the section IDs will be regenerated
        tags: List[str], optional
            A list of tags to add to the "tags" field of the item template dictionary
        extra_data: Dict[str, Any]
            Dictionary of data not associated with any field or section, for example login item URLs:
            {
                "urls": [
                    {
                        "label": "example website 2",
                        "href": "http://example2.website"
                    },
                    {
                        "label": "example website 1",
                        "primary": true,
                        "href": "https://example.website"
                    }
                ]
            }

        Raises
        ------
        OPSectionCollisionException
            If two or more sections have conflicting section IDs
        OPNewItemDataCollisionException
            If one or more keys in extra_data is already present in this item
        OPInvalidItemException
            If the subclass does not also inherit from a valid OPAbstractItem implementation
        """
        if fields is None:  # pragma: no coverage
            fields = []
        else:
            fields = list(fields)

        if sections is None:  # pragma: no coverage
            sections = []
        else:
            sections = list(sections)

        if tags is None:  # pragma: no coverage
            tags = []
        else:
            # remove duplicate tags
            tags = list(set(tags))

        if extra_data is None:  # pragma: no coverage
            extra_data = {}
        else:
            extra_data = dict(extra_data)

        directory = OPTemplateDirectory()

        # Mostly this satisfies mypy since OPNewItemMixin intentionally
        # doesn't define a CATEGORY attribute
        # my preference would be to not handle it and allow attribute error to blow
        # things up, but since we do have to check first, lets just raise
        # a descriptive OPInvalidItemException
        if hasattr(self, "CATEGORY"):
            template_dict: Dict = directory.template_for_category(
                self.CATEGORY)
        else:
            raise OPInvalidItemException(
                "Item template class must also from inherit a concreate OPAbstractItem implementation, overriding CATEGORY")
        template_dict["title"] = title
        section_map = {}
        new_sections = []
        new_fields = []
        old_to_new_sections = {}

        for sect in sections:
            if not isinstance(sect, OPNewSection):
                old_id = sect.section_id
                sect = OPNewSection.from_section(sect)
                # create a mapping of old-to-new sections to keep
                # up with the sections whose UUIDs get regenerated
                old_to_new_sections[old_id] = sect
            if sect.section_id in section_map:
                raise OPSectionCollisionException(
                    f"Duplicate section ID when creating new sections: {sect.section_id}")
            section_map[sect.section_id] = sect
            new_sections.append(sect)

        for field in fields:
            section_id = field.section_id
            if not isinstance(field, OPNewItemField):
                section = None
                if section_id:
                    section = old_to_new_sections.get(section_id)
                field = OPNewItemFieldFactory.item_field(
                    field, section=section)
            else:

                section = old_to_new_sections.get(
                    section_id) if section_id else None
                if section:
                    # If a section's UUID may have been regenerated, we need to re-link it to the field
                    field.update_section(sect)
            new_fields.append(field)
        template_dict["sections"] = new_sections
        template_dict["fields"] = new_fields
        template_dict["tags"] = tags
        key_collisions = []

        for key in extra_data.keys():
            if key in self:  # type: ignore
                #                    mypy doesn't like this even though it's fine later
                #                    and if it isn't we should ust blow up anyway
                key_collisions.append(key)  # pragma: no coverage
        if key_collisions:
            raise OPNewItemDataCollisionException(  # pragma: no coverage
                f"Extra data key collisions: {key_collisions}")
        if extra_data:
            template_dict.update(extra_data)

        # Satisfy mypy: Too many arguments for "__init__" of "object"
        args = [template_dict]
        super().__init__(*args)
        self._temp_files: List[str] = []

    def secure_tempfile(self, encoding="utf8") -> str:
        """
        Serialize this item object to a secure temp file for use during 'op item create'

        The temporary file is deleted when this object goes out of scope

        Parameters
        ----------
        encoding: str, optional
            Encoding to use when serializing to file. Defaults to "utf-8"

        Returns
        -------
        temp.name: str
            The name of the temporary file
        """
        temp = tempfile.NamedTemporaryFile(
            mode="w", delete=False, encoding=encoding)
        self._temp_files.append(temp.name)
        json.dump(self, temp)
        temp.close()
        return temp.name

    def supports_passwords(self) -> bool:
        """
        Whether passwords are applicable items of this type

        Returns
        -------
        password_supported: bool
        """
        return self.PASSWORDS_SUPPORTED

    def __del__(self):

        # if we blow up during object initialization
        # _temp_files may not exist, so check first
        if hasattr(self, "_temp_files"):
            while self._temp_files:
                t = self._temp_files.pop()
                try:
                    os.unlink(t)
                except FileNotFoundError:
                    continue



import copy
from typing import List, Union

from ...py_op_exceptions import OPInvalidItemException
from .item_field_base import OPItemField


class OPItemFieldCollisionException(Exception):
    pass


class OPSectionCollisionException(Exception):
    pass


class OPSection(dict):
    def __init__(self, section_dict):
        super().__init__(section_dict)
        # shadow fields map makes it easy to detect collisions
        # by looking up a field's ID to see if it's already been registered
        self._shadow_fields = {}

    @property
    def section_id(self) -> str:
        """
        Returns the section ID which may or may not be related to
        the user-visible title.
        It may be a lower-case transformation, like 'additional passwords'
        Or it may be something completely opaque, like
        'Section_967FEBAC931841BCBD2DD7CFE0B8DC82'
        """

        return self["id"]

    @property
    def label(self) -> str:
        """
        Returns the 'name' of the section as seen in the 1Password UI
        """
        return self["label"]

    @property
    def fields(self) -> List[OPItemField]:
        field_list = self.setdefault("fields", [])
        return field_list

    def register_field(self, field_dict: Union[OPItemField, dict], relaxed_validation: bool = False):
        if isinstance(field_dict, OPItemField):
            # make a copy of the field so we don't end up with a circular reference
            field = copy.copy(field_dict)
        else:
            field = OPItemField(field_dict)

        try:
            field_id = field.field_id
        except KeyError as e:
            # In theory there could be non-conformant data where
            # a field dictionary is missing an "ID" element
            # and we need to optionally be robust to that
            #
            # In practice this case should never occur
            # since the item object *should* raise an exception
            # before this method is ever called
            if not relaxed_validation:
                raise OPInvalidItemException(  # pragma: no coverage
                    f"Field has no ID {field.label}") from e
            else:
                # if relaxed validation is enabled, set field_id to empty string
                # since we have log below to handle that
                field_id = ""
        if field_id in self._shadow_fields:
            if not relaxed_validation:
                raise OPItemFieldCollisionException(
                    f"Field {field_id} already registered in section {self.section_id}")
            else:
                # for code coverage visibility
                pass
        else:
            self._shadow_fields[field_id] = field
        self.fields.append(field)

    def fields_by_label(self, label, case_sensitive=True) -> List[OPItemField]:
        """
        Returns all fields in a section matching the given label.
        Fields are not required to have unique labels, so there may be more than one match.
        """
        matching_fields = []
        f: OPItemField
        for f in self.fields:
            f_label = f.label
            if not case_sensitive:
                f_label = f_label.lower()
                label = label.lower()
            if f_label == label:
                matching_fields.append(f)
        return matching_fields

    def first_field_by_label(self, label: str, case_sensitive=True):
        """
        Convenience function for when you're certain there's only one field
        by a given label, or don't really care for whatever reason
        """
        fields = self.fields_by_label(label, case_sensitive=case_sensitive)
        f = fields[0]
        return f



from ..account import OPAccount
from ..op_items._item_list import OPItemList
from ..op_items._new_field_registry import OPNewItemField
from ..op_items._new_item import OPNewSection
from ..op_items.fields_sections._new_fields import (
    OPNewConcealedField,
    OPNewPasswordField,
    OPNewStringField,
    OPNewTOTPField,
    OPNewTOTPUri,
    OPNewUsernameField
)
from ..op_items.fields_sections.item_field import (
    OPConcealedField,
    OPStringField,
    OPTOTPField
)
from ..op_items.fields_sections.item_field_base import OPItemField
from ..op_items.fields_sections.item_section import OPSection
from ..op_items.item_types._item_base import OPAbstractItem
from ..op_items.item_types.api_credential import OPAPICredentialItem
from ..op_items.item_types.credit_card import OPCreditCardItem
from ..op_items.item_types.database import (
    OPDatabaseItem,
    OPDatabaseItemRelaxedValidation,
    OPDatabaseItemTemplate
)
from ..op_items.item_types.document import OPDocumentFile, OPDocumentItem
from ..op_items.item_types.login import (
    OPLoginItem,
    OPLoginItemNewPrimaryURL,
    OPLoginItemNewURL,
    OPLoginItemRelaxedValidation,
    OPLoginItemTemplate
)
from ..op_items.item_types.password import OPPasswordItem
from ..op_items.item_types.secure_note import OPSecureNoteItem
from ..op_items.item_types.server import (
    OPServerItem,
    OPServerItemRelaxedValidation
)
from ..op_items.item_types.ssh_key import OPSSHKeyItem
from ..op_items.password_recipe import OPPasswordRecipe
from ..op_items.totp import OPTOTPItem
from ..op_objects import OPGroup, OPUser, OPVault

# This causes these types to properly re-exported
# https://mypy.readthedocs.io/en/stable/config_file.html?highlight=export#confval-implicit_reexport
# anything that gets imported needs to be added to this list
__all__ = [
    "OPAccount",
    "OPItemList",
    "OPNewItemField",
    "OPNewConcealedField",
    "OPNewPasswordField",
    "OPNewStringField",
    "OPNewTOTPField",
    "OPNewTOTPUri",
    "OPNewUsernameField",
    "OPNewSection",
    "OPAbstractItem",
    "OPAPICredentialItem",
    "OPCreditCardItem",
    "OPDocumentFile",
    "OPDocumentItem",
    "OPConcealedField",
    "OPStringField",
    "OPTOTPField",
    "OPItemField",
    "OPSection",
    "OPLoginItem",
    "OPLoginItemNewPrimaryURL",
    "OPLoginItemNewURL",
    "OPLoginItemRelaxedValidation",
    "OPLoginItemTemplate",
    "OPDatabaseItem",
    "OPDatabaseItemRelaxedValidation",
    "OPDatabaseItemTemplate",
    "OPPasswordItem",
    "OPPasswordRecipe",
    "OPSecureNoteItem",
    "OPServerItem",
    "OPServerItemRelaxedValidation",
    "OPSSHKeyItem",
    "OPTOTPItem",
    "OPGroup",
    "OPUser",
    "OPVault"
]
